# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import concurrent.futures
import copy
import functools
import queue
import types
import typing as ty
import warnings
import weakref

import dogpile.cache
import keystoneauth1.exceptions
from keystoneauth1.identity import base as ks_plugin_base
import requests.models
import requestsexceptions
import typing_extensions as ty_ext

from openstack import _log
from openstack import _services_mixin
from openstack.cloud import _utils
from openstack.cloud import meta
from openstack import config as cloud_config
from openstack.config import cloud_region
from openstack import exceptions
from openstack import proxy
from openstack import resource
from openstack import utils
from openstack import warnings as os_warnings

if ty.TYPE_CHECKING:
    from dogpile.cache import region as cache_region
    from keystoneauth1.access import service_catalog as ks_service_catalog
    from keystoneauth1 import session as ks_session
    from oslo_config import cfg

    from openstack import service_description


class _OpenStackCloudMixin(_services_mixin.ServicesMixin):
    """Represent a connection to an OpenStack Cloud.

    OpenStackCloud is the entry point for all cloud operations, regardless
    of which OpenStack service those operations may ultimately come from.
    The operations on an OpenStackCloud are resource oriented rather than
    REST API operation oriented. For instance, one will request a Floating IP
    and that Floating IP will be actualized either via neutron or via nova
    depending on how this particular cloud has decided to arrange itself.
    """

    _OBJECT_MD5_KEY = 'x-sdk-md5'
    _OBJECT_SHA256_KEY = 'x-sdk-sha256'
    _OBJECT_AUTOCREATE_KEY = 'x-sdk-autocreated'
    _OBJECT_AUTOCREATE_CONTAINER = 'images'

    # NOTE(shade) shade keys were x-object-meta-x-shade-md5 - we need to check
    #             those in freshness checks so that a shade->sdk transition
    #             doesn't result in a re-upload
    _SHADE_OBJECT_MD5_KEY = 'x-object-meta-x-shade-md5'
    _SHADE_OBJECT_SHA256_KEY = 'x-object-meta-x-shade-sha256'
    _SHADE_OBJECT_AUTOCREATE_KEY = 'x-object-meta-x-shade-autocreated'

    config: cloud_region.CloudRegion

    def __init__(
        self,
        cloud: str | None = None,
        config: cloud_region.CloudRegion | None = None,
        session: ty.Optional['ks_session.Session'] = None,
        app_name: str | None = None,
        app_version: str | None = None,
        extra_services: list['service_description.ServiceDescription']
        | None = None,
        strict: bool = False,
        use_direct_get: bool | None = None,
        task_manager: ty.Any = None,
        rate_limit: float | dict[str, float] | None = None,
        oslo_conf: ty.Optional['cfg.ConfigOpts'] = None,
        service_types: list[str] | None = None,
        global_request_id: str | None = None,
        strict_proxies: bool = False,
        pool_executor: concurrent.futures.Executor | None = None,
        **kwargs: ty.Any,
    ) -> None:
        """Create a connection to a cloud.

        A connection needs information about how to connect, how to
        authenticate and how to select the appropriate services to use.

        The recommended way to provide this information is by referencing
        a named cloud config from an existing `clouds.yaml` file. The cloud
        name ``envvars`` may be used to consume a cloud configured via ``OS_``
        environment variables.

        A pre-existing :class:`~openstack.config.cloud_region.CloudRegion`
        object can be passed in lieu of a cloud name, for cases where the user
        already has a fully formed CloudRegion and just wants to use it.

        Similarly, if for some reason the user already has a
        :class:`~keystoneauth1.session.Session` and wants to use it, it may be
        passed in.

        :param str cloud: Name of the cloud from config to use.
        :param config: CloudRegion object representing the config for the
            region of the cloud in question.
        :type config: :class:`~openstack.config.cloud_region.CloudRegion`
        :param session: A session object compatible with
            :class:`~keystoneauth1.session.Session`.
        :type session: :class:`~keystoneauth1.session.Session`
        :param str app_name: Name of the application to be added to User Agent.
        :param str app_version: Version of the application to be added to
            User Agent.
        :param extra_services: List of
            :class:`~openstack.service_description.ServiceDescription`
            objects describing services that openstacksdk otherwise does not
            know about.
        :param bool use_direct_get:
            For get methods, make specific REST calls for server-side
            filtering instead of making list calls and filtering client-side.
            Default false.
        :param task_manager:
            Ignored. Exists for backwards compat during transition. Rate limit
            parameters should be passed directly to the `rate_limit` parameter.
        :param rate_limit:
            Client-side rate limit, expressed in calls per second. The
            parameter can either be a single float, or it can be a dict with
            keys as service-type and values as floats expressing the calls
            per second for that service. Defaults to None, which means no
            rate-limiting is performed.
        :param oslo_conf: An oslo.config CONF object.
        :type oslo_conf: :class:`~oslo_config.cfg.ConfigOpts`
            An oslo.config ``CONF`` object that has been populated with
            ``keystoneauth1.loading.register_adapter_conf_options`` in
            groups named by the OpenStack service's project name.
        :param service_types:
            A list/set of service types this Connection should support. All
            other service types will be disabled (will error if used).
            **Currently only supported in conjunction with the ``oslo_conf``
            kwarg.**
        :param global_request_id: A Request-id to send with all interactions.
        :param strict_proxies:
            If True, check proxies on creation and raise
            ServiceDiscoveryException if the service is unavailable.
        :type strict_proxies: bool
            Throw an ``openstack.exceptions.ServiceDiscoveryException`` if the
            endpoint for a given service doesn't work. This is useful for
            OpenStack services using sdk to talk to other OpenStack services
            where it can be expected that the deployer config is correct and
            errors should be reported immediately.
            Default false.
        :param pool_executor:
        :type pool_executor: :class:`~futurist.Executor`
            A futurist ``Executor`` object to be used for concurrent background
            activities. Defaults to None in which case a ThreadPoolExecutor
            will be created if needed.
        :param kwargs: If a config is not provided, the rest of the parameters
            provided are assumed to be arguments to be passed to the
            CloudRegion constructor.
        """
        super().__init__()

        if use_direct_get is not None:
            warnings.warn(
                "The 'use_direct_get' argument is deprecated for removal",
                os_warnings.RemovedInSDK50Warning,
            )

        self._extra_services = {}
        self._strict_proxies = strict_proxies
        if extra_services:
            for service in extra_services:
                self._extra_services[service.service_type] = service

        if config:
            self.config = config
        else:
            if oslo_conf:
                self.config = cloud_region.from_conf(
                    oslo_conf,
                    session=session,
                    app_name=app_name,
                    app_version=app_version,
                    service_types=service_types,
                )
            elif session:
                self.config = cloud_region.from_session(
                    session=session,
                    app_name=app_name,
                    app_version=app_version,
                    load_yaml_config=False,
                    load_envvars=False,
                    rate_limit=rate_limit,
                    **kwargs,
                )
            else:
                self.config = cloud_config.get_cloud_region(
                    cloud=cloud,
                    app_name=app_name,
                    app_version=app_version,
                    load_yaml_config=cloud is not None,
                    load_envvars=cloud is not None,
                    rate_limit=rate_limit,
                    **kwargs,
                )

        self._session: ks_session.Session | None = None
        self._proxies: dict[str, proxy.Proxy] = {}
        self.__pool_executor = pool_executor
        self._global_request_id = global_request_id
        self.use_direct_get = use_direct_get or False
        self.strict_mode = strict

        self.log = _log.setup_logging('openstack')

        self.name = self.config.name
        self.auth = self.config.get_auth_args()
        self.default_interface = self.config.get_interface()
        self.force_ipv4 = self.config.force_ipv4

        (self.verify, self.cert) = self.config.get_requests_verify_args()

        # Turn off urllib3 warnings about insecure certs if we have
        # explicitly configured requests to tell it we do not want
        # cert verification
        if not self.verify:
            self.log.debug(
                "Turning off Insecure SSL warnings since verify=False"
            )
            category = requestsexceptions.InsecureRequestWarning
            if category:
                # InsecureRequestWarning references a Warning class or is None
                warnings.filterwarnings('ignore', category=category)

        self._disable_warnings: dict[str, bool] = {}

        cache_expiration_time = int(self.config.get_cache_expiration_time())
        cache_class = self.config.get_cache_class()
        cache_arguments = self.config.get_cache_arguments()

        self._cache_expirations = dict()

        if cache_class != 'dogpile.cache.null':
            self.cache_enabled = True
        else:
            self.cache_enabled = False

        # Uncoditionally create cache even with a "null" backend
        self._cache = self._make_cache(
            cache_class, cache_expiration_time, cache_arguments
        )
        expirations = self.config.get_cache_expirations()
        for expire_key in expirations.keys():
            self._cache_expirations[expire_key] = expirations[expire_key]

        self._api_cache_keys: set[str] = set()

        self._local_ipv6 = (
            _utils.localhost_supports_ipv6() if not self.force_ipv4 else False
        )

        # Register cleanup steps
        atexit.register(self.close)

    @property
    def session(self) -> 'ks_session.Session':
        if not self._session:
            self._session = self.config.get_session()
            # Hide a reference to the connection on the session to help with
            # backwards compatibility for folks trying to just pass
            # conn.session to a Resource method's session argument.
            setattr(self.session, '_sdk_connection', weakref.proxy(self))
        return self._session

    @property
    def _pool_executor(self) -> concurrent.futures.Executor:
        if not self.__pool_executor:
            self.__pool_executor = concurrent.futures.ThreadPoolExecutor(
                max_workers=5
            )
        return self.__pool_executor

    def close(self) -> None:
        """Release any resources held open."""
        self.config.set_auth_cache()
        if self.__pool_executor:
            self.__pool_executor.shutdown()
        atexit.unregister(self.close)

    def __enter__(self) -> ty_ext.Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        self.close()

    def set_global_request_id(self, global_request_id: str) -> None:
        self._global_request_id = global_request_id

    def global_request(self, global_request_id: str) -> ty_ext.Self:
        """Make a new Connection object with a global request id set.

        Take the existing settings from the current Connection and construct a
        new Connection object with the global_request_id overridden.

        .. code-block:: python

            from oslo_context import context

            cloud = openstack.connect(cloud='example')
            # Work normally
            servers = cloud.list_servers()
            cloud2 = cloud.global_request(context.generate_request_id())
            # cloud2 sends all requests with global_request_id set
            servers = cloud2.list_servers()

        Additionally, this can be used as a context manager:

        .. code-block:: python

            from oslo_context import context

            c = openstack.connect(cloud='example')
            # Work normally
            servers = c.list_servers()
            with c.global_request(context.generate_request_id()) as c2:
                # c2 sends all requests with global_request_id set
                servers = c2.list_servers()

        :param global_request_id: The `global_request_id` to send.
        """
        params = copy.deepcopy(self.config.config)
        config = cloud_region.from_session(
            session=self.session,
            app_name=self.config._app_name,
            app_version=self.config._app_version,
            discovery_cache=self.session._discovery_cache,
            **params,
        )

        # Override the cloud name so that logging/location work right
        config._name = self.name
        config.config['profile'] = self.name
        # Use self.__class__ so that we return whatever this is, like if it's
        # a subclass in the case of shade wrapping sdk.
        new_conn = self.__class__(config=config)
        new_conn.set_global_request_id(global_request_id)
        return new_conn

    def _make_cache(
        self,
        cache_class: str,
        expiration_time: int,
        arguments: dict[str, ty.Any] | None,
    ) -> 'cache_region.CacheRegion':
        return dogpile.cache.make_region(
            function_key_generator=self._make_cache_key
        ).configure(
            cache_class, expiration_time=expiration_time, arguments=arguments
        )

    def _make_cache_key(
        self, namespace: str, fn: ty.Callable[..., ty.Any]
    ) -> ty.Callable[..., str]:
        fname = fn.__name__
        if namespace is None:
            name_key = self.name
        else:
            name_key = f'{self.name}:{namespace}'

        def generate_key(*args, **kwargs):
            # TODO(frickler): make handling arg keys actually work
            arg_key = ''
            kw_keys = sorted(kwargs.keys())
            kwargs_key = ','.join(
                [f'{k}:{kwargs[k]}' for k in kw_keys if k != 'cache']
            )
            ans = "_".join([str(name_key), fname, arg_key, kwargs_key])
            return ans

        return generate_key

    def pprint(self, resource: object) -> None:
        """Wrapper around pprint that groks munch objects"""
        # import late since this is a utility function
        import pprint

        new_resource = _utils._dictify_resource(resource)
        pprint.pprint(new_resource)

    def pformat(self, resource: object) -> str:
        """Wrapper around pformat that groks munch objects"""
        # import late since this is a utility function
        import pprint

        new_resource = _utils._dictify_resource(resource)
        return pprint.pformat(new_resource)

    @property
    def _keystone_catalog(self) -> 'ks_service_catalog.ServiceCatalog':
        if self.session.auth is None:
            raise exceptions.ConfigException(
                'session has no auth information attached'
            )

        if not isinstance(
            self.session.auth, ks_plugin_base.BaseIdentityPlugin
        ):
            raise exceptions.ConfigException(
                'cannot fetch catalog for non-keystone auth plugin'
            )

        return self.session.auth.get_access(self.session).service_catalog

    @property
    def service_catalog(self) -> list[dict[str, ty.Any]]:
        return self._keystone_catalog.catalog

    @property
    def auth_token(self) -> str | None:
        # Keystone's session will reuse a token if it is still valid.
        # We don't need to track validity here, just get_token() each time.
        return self.session.get_token()

    @property
    def current_user_id(self) -> str | None:
        """Get the id of the currently logged-in user from the token."""
        if self.session.auth is None:
            raise exceptions.ConfigException(
                'session has no auth information attached'
            )

        if not isinstance(
            self.session.auth, ks_plugin_base.BaseIdentityPlugin
        ):
            raise exceptions.ConfigException(
                'cannot fetch catalog for non-keystone auth plugin'
            )

        return self.session.auth.get_access(self.session).user_id

    @property
    def current_project_id(self) -> str | None:
        """Get the current project ID.

        Returns the project_id of the current token scope. None means that
        the token is domain scoped or unscoped.

        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
            if a new token fetch fails.
        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
            if a plugin is not available.
        """
        return self.session.get_project_id()

    @property
    def current_project(self) -> utils.Munch:
        """Return a ``utils.Munch`` describing the current project"""
        return self._get_project_info()

    def _get_project_info(self, project_id: str | None = None) -> utils.Munch:
        project_info = utils.Munch(
            id=project_id,
            name=None,
            domain_id=None,
            domain_name=None,
        )
        if not project_id or project_id == self.current_project_id:
            # If we don't have a project_id parameter, it means a user is
            # directly asking what the current state is.
            # Alternately, if we have one, that means we're calling this
            # from within a normalize function, which means the object has
            # a project_id associated with it. If the project_id matches
            # the project_id of our current token, that means we can supplement
            # the info with human readable info about names if we have them.
            # If they don't match, that means we're an admin who has pulled
            # an object from a different project, so adding info from the
            # current token would be wrong.
            auth_args = self.config.config.get('auth', {})
            project_info['id'] = self.current_project_id
            project_info['name'] = auth_args.get('project_name')
            project_info['domain_id'] = auth_args.get('project_domain_id')
            project_info['domain_name'] = auth_args.get('project_domain_name')
        return project_info

    @property
    def current_location(self) -> utils.Munch:
        """Return a ``utils.Munch`` explaining the current cloud location."""
        return self._get_current_location()

    def _get_current_location(
        self,
        project_id: str | None = None,
        zone: str | None = None,
    ) -> utils.Munch:
        return utils.Munch(
            cloud=self.name,
            # TODO(efried): This is wrong, but it only seems to be used in a
            # repr; can we get rid of it?
            region_name=self.config.get_region_name(),
            zone=zone,
            project=self._get_project_info(project_id),
        )

    def range_search(self, data, filters):
        """Perform integer range searches across a list of dictionaries.

        Given a list of dictionaries, search across the list using the given
        dictionary keys and a range of integer values for each key. Only
        dictionaries that match ALL search filters across the entire original
        data set will be returned.

        It is not a requirement that each dictionary contain the key used
        for searching. Those without the key will be considered non-matching.

        The range values must be string values and is either a set of digits
        representing an integer for matching, or a range operator followed by
        a set of digits representing an integer for matching. If a range
        operator is not given, exact value matching will be used. Valid
        operators are one of: <,>,<=,>=

        :param data: List of dictionaries to be searched.
        :param filters: Dict describing the one or more range searches to
            perform. If more than one search is given, the result will be the
            members of the original data set that match ALL searches. An
            example of filtering by multiple ranges::

                {"vcpus": "<=5", "ram": "<=2048", "disk": "1"}

        :returns: A list subset of the original data set.
        :raises: :class:`~openstack.exceptions.SDKException` on invalid range
            expressions.
        """
        filtered: list[object] = []

        for key, range_value in filters.items():
            # We always want to operate on the full data set so that
            # calculations for minimum and maximum are correct.
            results = _utils.range_filter(data, key, range_value)

            if not filtered:
                # First set of results
                filtered = results
            else:
                # The combination of all searches should be the intersection of
                # all result sets from each search. So adjust the current set
                # of filtered data by computing its intersection with the
                # latest result set.
                filtered = [r for r in results for f in filtered if r == f]

        return filtered

    def _get_and_munchify(self, key, data):
        """Wrapper around meta.get_and_munchify.

        Some of the methods expect a `meta` attribute to be passed in as
        part of the method signature. In those methods the meta param is
        overriding the meta module making the call to meta.get_and_munchify
        to fail.
        """
        if isinstance(data, requests.models.Response):
            data = proxy._json_response(data)
        return meta.get_and_munchify(key, data)

    def get_name(self) -> str:
        return self.name

    def get_session_endpoint(
        self,
        service_key: str,
        min_version: str | None = None,
        max_version: str | None = None,
    ) -> str | None:
        try:
            return self.config.get_session_endpoint(
                service_key,
                min_version=min_version,
                max_version=max_version,
            )
        except keystoneauth1.exceptions.catalog.EndpointNotFound as e:
            self.log.debug(
                "Endpoint not found in %s cloud: %s", self.name, str(e)
            )
            endpoint = None
        except exceptions.SDKException:
            raise
        except Exception as e:
            raise exceptions.SDKException(
                f"Error getting {service_key} endpoint on {self.name}:{self.config.get_region_name(service_key)}: "
                f"{str(e)}"
            )
        return endpoint

    def has_service(
        self, service_key: str, version: str | None = None
    ) -> bool:
        if not self.config.has_service(service_key):
            # TODO(mordred) add a stamp here so that we only report this once
            if not (
                service_key in self._disable_warnings
                and self._disable_warnings[service_key]
            ):
                self.log.debug(
                    "Disabling %(service_key)s entry in catalog per config",
                    {'service_key': service_key},
                )
                self._disable_warnings[service_key] = True
            return False
        try:
            kwargs = dict()
            # If a specific version was requested - try it
            if version is not None:
                kwargs['min_version'] = version
                kwargs['max_version'] = version
            endpoint = self.get_session_endpoint(service_key, **kwargs)
        except exceptions.SDKException:
            return False
        if endpoint:
            return True
        else:
            return False

    def search_resources(
        self,
        resource_type,
        name_or_id,
        get_args=None,
        get_kwargs=None,
        list_args=None,
        list_kwargs=None,
        **filters,
    ):
        """Search resources

        Search resources matching certain conditions

        :param str resource_type: String representation of the expected
            resource as `service.resource` (i.e. "network.security_group").
        :param str name_or_id: Name or ID of the resource
        :param list get_args: Optional args to be passed to the _get call.
        :param dict get_kwargs: Optional kwargs to be passed to the _get call.
        :param list list_args: Optional args to be passed to the _list call.
        :param dict list_kwargs: Optional kwargs to be passed to the _list call
        :param dict filters: Additional filters to be used for querying
            resources.
        """
        get_args = get_args or ()
        get_kwargs = get_kwargs or {}
        list_args = list_args or ()
        list_kwargs = list_kwargs or {}

        # User used string notation. Try to find proper
        # resource
        (service_name, resource_name) = resource_type.split('.')
        if not hasattr(self, service_name):
            raise exceptions.SDKException(
                f"service {service_name} is not existing/enabled"
            )
        service_proxy = getattr(self, service_name)
        try:
            resource_type = service_proxy._resource_registry[resource_name]
        except KeyError:
            raise exceptions.SDKException(
                f"Resource {resource_name} is not known in service {service_name}"
            )

        if name_or_id:
            # name_or_id is definitely not None
            try:
                resource_by_id = service_proxy._get(
                    resource_type, name_or_id, *get_args, **get_kwargs
                )
                return [resource_by_id]
            except exceptions.NotFoundException:
                pass

        if not filters:
            filters = {}

        if name_or_id:
            filters["name"] = name_or_id
        list_kwargs.update(filters)

        return list(
            service_proxy._list(resource_type, *list_args, **list_kwargs)
        )

    def project_cleanup(
        self,
        dry_run=True,
        wait_timeout=120,
        status_queue=None,
        filters=None,
        resource_evaluation_fn=None,
        skip_resources=None,
    ):
        """Cleanup the project resources.

        Cleanup all resources in all services, which provide cleanup methods.

        :param bool dry_run: Cleanup or only list identified resources.
        :param int wait_timeout: Maximum amount of time given to each service
            to comlete the cleanup.
        :param queue status_queue: a threading queue object used to get current
            process status. The queue contain processed resources.
        :param dict filters: Additional filters for the cleanup (only resources
            matching all filters will be deleted, if there are no other
            dependencies).
        :param resource_evaluation_fn: A callback function, which will be
            invoked for each resurce and must return True/False depending on
            whether resource need to be deleted or not.
        :param skip_resources: List of specific resources whose cleanup should
            be skipped.
        """
        dependencies = {}
        get_dep_fn_name = '_get_cleanup_dependencies'
        cleanup_fn_name = '_service_cleanup'
        if not status_queue:
            status_queue = queue.Queue()
        for service in self.config.get_enabled_services():
            try:
                if hasattr(self, service):
                    proxy = getattr(self, service)
                    if (
                        proxy
                        and hasattr(proxy, get_dep_fn_name)
                        and hasattr(proxy, cleanup_fn_name)
                    ):
                        deps = getattr(proxy, get_dep_fn_name)()
                        if deps:
                            dependencies.update(deps)
            except (
                exceptions.NotSupported,
                exceptions.ServiceDisabledException,
            ):
                # Cloud may include endpoint in catalog but not
                # implement the service or disable it
                pass
        dep_graph = utils.TinyDAG()
        for k, v in dependencies.items():
            dep_graph.add_node(k)
            for dep in v['before']:
                dep_graph.add_node(dep)
                dep_graph.add_edge(k, dep)
            for dep in v.get('after', []):
                dep_graph.add_edge(dep, k)

        cleanup_resources: dict[str, resource.Resource] = {}

        for service in dep_graph.walk(timeout=wait_timeout):
            fn = None
            try:
                if hasattr(self, service):
                    proxy = getattr(self, service)
                    cleanup_fn = getattr(proxy, cleanup_fn_name, None)
                    if cleanup_fn:
                        fn = functools.partial(
                            cleanup_fn,
                            dry_run=dry_run,
                            client_status_queue=status_queue,
                            identified_resources=cleanup_resources,
                            filters=filters,
                            resource_evaluation_fn=resource_evaluation_fn,
                            skip_resources=skip_resources,
                        )
            except exceptions.ServiceDisabledException:
                # same reason as above
                pass
            if fn:
                self._pool_executor.submit(
                    cleanup_task, dep_graph, service, fn
                )
            else:
                dep_graph.node_done(service)

        for count in utils.iterate_timeout(
            timeout=wait_timeout,
            message="Timeout waiting for cleanup to finish",
            wait=1,
        ):
            if dep_graph.is_complete():
                return


def cleanup_task(graph, service, fn):
    try:
        fn()
    except Exception:
        log = _log.setup_logging('openstack.project_cleanup')
        log.exception(f'Error in the {service} cleanup function')
    finally:
        graph.node_done(service)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from urllib import parse

from openstack import resource


class Template(resource.Resource):
    # capabilities
    allow_create = False
    allow_list = False
    allow_fetch = False
    allow_delete = False
    allow_commit = False

    # Properties
    #: The description specified in the template
    description = resource.Body('Description')
    #: Key and value pairs that contain template parameters
    parameters = resource.Body('Parameters', type=dict)
    #: A list of parameter groups each contains a lsit of parameter names.
    parameter_groups = resource.Body('ParameterGroups', type=list)

    def validate(
        self,
        session,
        template,
        environment=None,
        template_url=None,
        ignore_errors=None,
    ):
        url = '/validate'

        body = {'template': template}
        if environment is not None:
            body['environment'] = environment
        if template_url is not None:
            body['template_url'] = template_url
        if ignore_errors:
            qry = parse.urlencode({'ignore_errors': ignore_errors})
            url = '?'.join([url, qry])

        resp = session.post(url, json=body)
        self._translate_response(resp)
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.tests import fakes
from openstack.tests.unit import base


class TestServerConsole(base.TestCase):
    def setUp(self):
        super().setUp()

        self.server_id = str(uuid.uuid4())
        self.server_name = self.getUniqueString('name')
        self.server = fakes.make_fake_server(
            server_id=self.server_id, name=self.server_name
        )
        self.output = self.getUniqueString('output')

    def test_get_server_console_dict(self):
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/servers/{self.server_id}/action',
                    json={"output": self.output},
                    validate=dict(json={'os-getConsoleOutput': {'length': 5}}),
                ),
            ]
        )

        self.assertEqual(
            self.output, self.cloud.get_server_console(self.server, 5)
        )
        self.assert_calls()

    def test_get_server_console_name_or_id(self):
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/servers/{self.server_id}',
                    json={'server': self.server},
                ),
                dict(
                    method='POST',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/servers/{self.server_id}/action',
                    json={"output": self.output},
                    validate=dict(json={'os-getConsoleOutput': {}}),
                ),
            ]
        )

        self.assertEqual(
            self.output, self.cloud.get_server_console(self.server['id'])
        )

        self.assert_calls()

    def test_get_server_console_no_console(self):
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=f'{fakes.COMPUTE_ENDPOINT}/servers/{self.server_id}/action',
                    status_code=400,
                    validate=dict(json={'os-getConsoleOutput': {}}),
                ),
            ]
        )

        self.assertEqual('', self.cloud.get_server_console(self.server))

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import vpn_ipsec_site_connection
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    "admin_state_up": True,
    "auth_mode": "1",
    "ikepolicy_id": "2",
    "vpnservice_id": "3",
    "local_ep_group_id": "4",
    "peer_address": "5",
    "route_mode": "6",
    "ipsecpolicy_id": "7",
    "peer_id": "8",
    "psk": "9",
    "description": "10",
    "initiator": "11",
    "peer_cidrs": ['1', '2'],
    "name": "12",
    "tenant_id": "13",
    "interval": 5,
    "mtu": 5,
    "peer_ep_group_id": "14",
    "dpd": {'a': 5},
    "timeout": 16,
    "action": "17",
    "local_id": "18",
}


class TestVpnIPSecSiteConnection(base.TestCase):
    def test_basic(self):
        sot = vpn_ipsec_site_connection.VpnIPSecSiteConnection()
        self.assertEqual('ipsec_site_connection', sot.resource_key)
        self.assertEqual('ipsec_site_connections', sot.resources_key)
        self.assertEqual('/vpn/ipsec-site-connections', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = vpn_ipsec_site_connection.VpnIPSecSiteConnection(**EXAMPLE)
        self.assertTrue(sot.is_admin_state_up)
        self.assertEqual(EXAMPLE['auth_mode'], sot.auth_mode)
        self.assertEqual(EXAMPLE['ikepolicy_id'], sot.ikepolicy_id)
        self.assertEqual(EXAMPLE['vpnservice_id'], sot.vpnservice_id)
        self.assertEqual(EXAMPLE['local_ep_group_id'], sot.local_ep_group_id)
        self.assertEqual(EXAMPLE['peer_address'], sot.peer_address)
        self.assertEqual(EXAMPLE['route_mode'], sot.route_mode)
        self.assertEqual(EXAMPLE['ipsecpolicy_id'], sot.ipsecpolicy_id)
        self.assertEqual(EXAMPLE['peer_id'], sot.peer_id)
        self.assertEqual(EXAMPLE['psk'], sot.psk)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertEqual(EXAMPLE['initiator'], sot.initiator)
        self.assertEqual(EXAMPLE['peer_cidrs'], sot.peer_cidrs)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['tenant_id'], sot.project_id)
        self.assertEqual(EXAMPLE['interval'], sot.interval)
        self.assertEqual(EXAMPLE['mtu'], sot.mtu)
        self.assertEqual(EXAMPLE['peer_ep_group_id'], sot.peer_ep_group_id)
        self.assertEqual(EXAMPLE['dpd'], sot.dpd)
        self.assertEqual(EXAMPLE['timeout'], sot.timeout)
        self.assertEqual(EXAMPLE['action'], sot.action)
        self.assertEqual(EXAMPLE['local_id'], sot.local_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_security_groups
----------------------------------

Functional tests for security_groups resource.
"""

from openstack.tests.functional import base


class TestSecurityGroups(base.BaseFunctionalTest):
    def test_create_list_security_groups(self):
        sg1 = self.user_cloud.create_security_group(
            name="sg1", description="sg1"
        )
        self.addCleanup(self.user_cloud.delete_security_group, sg1['id'])
        if self.user_cloud.has_service('network'):
            # Neutron defaults to all_tenants=1 when admin
            sg_list = self.user_cloud.list_security_groups()
            self.assertIn(sg1['id'], [sg['id'] for sg in sg_list])

        else:
            # Nova does not list all tenants by default
            sg_list = self.operator_cloud.list_security_groups()

    def test_create_list_security_groups_operator(self):
        if not self.operator_cloud:
            self.skipTest("Operator cloud is required for this test")

        sg1 = self.user_cloud.create_security_group(
            name="sg1", description="sg1"
        )
        self.addCleanup(self.user_cloud.delete_security_group, sg1['id'])
        sg2 = self.operator_cloud.create_security_group(
            name="sg2", description="sg2"
        )
        self.addCleanup(self.operator_cloud.delete_security_group, sg2['id'])

        if self.user_cloud.has_service('network'):
            # Neutron defaults to all_tenants=1 when admin
            sg_list = self.operator_cloud.list_security_groups()
            self.assertIn(sg1['id'], [sg['id'] for sg in sg_list])

            # Filter by tenant_id (filtering by project_id won't work with
            # Keystone V2)
            sg_list = self.operator_cloud.list_security_groups(
                filters={'tenant_id': self.user_cloud.current_project_id}
            )
            self.assertIn(sg1['id'], [sg['id'] for sg in sg_list])
            self.assertNotIn(sg2['id'], [sg['id'] for sg in sg_list])

        else:
            # Nova does not list all tenants by default
            sg_list = self.operator_cloud.list_security_groups()
            self.assertIn(sg2['id'], [sg['id'] for sg in sg_list])
            self.assertNotIn(sg1['id'], [sg['id'] for sg in sg_list])

            sg_list = self.operator_cloud.list_security_groups(
                filters={'all_tenants': 1}
            )
            self.assertIn(sg1['id'], [sg['id'] for sg in sg_list])

    def test_get_security_group_by_id(self):
        sg = self.user_cloud.create_security_group(name='sg', description='sg')
        self.addCleanup(self.user_cloud.delete_security_group, sg['id'])

        ret_sg = self.user_cloud.get_security_group_by_id(sg['id'])
        self.assertEqual(sg, ret_sg)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.identity.v3 import trust
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'allow_redelegation': False,
    'expires_at': '2016-03-09T12:14:57.233772',
    'id': IDENTIFIER,
    'impersonation': True,
    'links': {'self': 'fake_link'},
    'project_id': '1',
    'redelegated_trust_id': None,
    'redelegation_count': '0',
    'remaining_uses': 10,
    'role_links': {'self': 'other_fake_link'},
    'trustee_user_id': '2',
    'trustor_user_id': '3',
    'roles': [{'name': 'test-role'}],
}


class TestTrust(base.TestCase):
    def test_basic(self):
        sot = trust.Trust()
        self.assertEqual('trust', sot.resource_key)
        self.assertEqual('trusts', sot.resources_key)
        self.assertEqual('/OS-TRUST/trusts', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = trust.Trust(**EXAMPLE)
        self.assertEqual(EXAMPLE['allow_redelegation'], sot.allow_redelegation)
        self.assertEqual(EXAMPLE['expires_at'], sot.expires_at)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertTrue(sot.is_impersonation)
        self.assertEqual(EXAMPLE['links'], sot.links)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['role_links'], sot.role_links)
        self.assertEqual(
            EXAMPLE['redelegated_trust_id'], sot.redelegated_trust_id
        )
        self.assertEqual(EXAMPLE['remaining_uses'], sot.remaining_uses)
        self.assertEqual(EXAMPLE['trustee_user_id'], sot.trustee_user_id)
        self.assertEqual(EXAMPLE['trustor_user_id'], sot.trustor_user_id)
        self.assertEqual(EXAMPLE['roles'], sot.roles)
        self.assertEqual(EXAMPLE['redelegation_count'], sot.redelegation_count)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.container_infrastructure_management.v1 import cluster
from openstack.tests.unit import base

EXAMPLE = {
    "cluster_template_id": "0562d357-8641-4759-8fed-8173f02c9633",
    "create_timeout": 60,
    "discovery_url": None,
    "flavor_id": None,
    "keypair": "my_keypair",
    "labels": {},
    "master_count": 2,
    "master_flavor_id": None,
    "name": "k8s",
    "node_count": 2,
}


class TestCluster(base.TestCase):
    def test_basic(self):
        sot = cluster.Cluster()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('clusters', sot.resources_key)
        self.assertEqual('/clusters', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = cluster.Cluster(**EXAMPLE)
        self.assertEqual(
            EXAMPLE['cluster_template_id'],
            sot.cluster_template_id,
        )
        self.assertEqual(EXAMPLE['create_timeout'], sot.create_timeout)
        self.assertEqual(EXAMPLE['discovery_url'], sot.discovery_url)
        self.assertEqual(EXAMPLE['flavor_id'], sot.flavor_id)
        self.assertEqual(EXAMPLE['keypair'], sot.keypair)
        self.assertEqual(EXAMPLE['labels'], sot.labels)
        self.assertEqual(EXAMPLE['master_count'], sot.master_count)
        self.assertEqual(EXAMPLE['master_flavor_id'], sot.master_flavor_id)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['node_count'], sot.node_count)



# Copyright 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import operator
import tempfile
from unittest import mock
import uuid

from openstack.cloud import meta
from openstack import exceptions
from openstack.image.v1 import image as image_v1
from openstack.image.v2 import image
from openstack.tests import fakes
from openstack.tests.unit import base


IMPORT_METHODS = 'glance-direct,web-download'


class BaseTestImage(base.TestCase):
    def setUp(self):
        super().setUp()
        self.image_id = str(uuid.uuid4())
        self.image_name = self.getUniqueString('image')
        self.object_name = f'images/{self.image_name}'
        self.imagefile = tempfile.NamedTemporaryFile(delete=False)
        data = b'\2\0'
        self.imagefile.write(data)
        self.imagefile.close()
        self.output = data
        self.fake_image_dict = fakes.make_fake_image(
            image_id=self.image_id,
            image_name=self.image_name,
            data=self.imagefile.name,
        )
        self.fake_search_return = {'images': [self.fake_image_dict]}
        self.container_name = self.getUniqueString('container')

    def _compare_images(self, exp, real):
        self.assertDictEqual(
            image.Image(**exp).to_dict(computed=False),
            real.to_dict(computed=False),
        )

    def _compare_images_v1(self, exp, real):
        self.assertDictEqual(
            image_v1.Image(**exp).to_dict(computed=False),
            real.to_dict(computed=False),
        )


class TestImage(BaseTestImage):
    def setUp(self):
        super().setUp()
        self.use_glance()

    def test_download_image_no_output(self):
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.download_image,
            self.image_name,
        )

    def test_download_image_two_outputs(self):
        fake_fd = io.BytesIO()
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.download_image,
            self.image_name,
            output_path='fake_path',
            output_file=fake_fd,
        )

    def test_download_image_no_images_found(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'https://image.example.com/v2/images/{self.image_name}',
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=f'https://image.example.com/v2/images?name={self.image_name}',  # noqa: E501
                    json=dict(images=[]),
                ),
                dict(
                    method='GET',
                    uri='https://image.example.com/v2/images?os_hidden=True',
                    json=dict(images=[]),
                ),
            ]
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.cloud.download_image,
            self.image_name,
            output_path='fake_path',
        )
        self.assert_calls()

    def _register_image_mocks(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=f'https://image.example.com/v2/images/{self.image_name}',
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=f'https://image.example.com/v2/images?name={self.image_name}',  # noqa: E501
                    json=self.fake_search_return,
                ),
                dict(
                    method='GET',
                    uri=f'https://image.example.com/v2/images/{self.image_id}/file',
                    content=self.output,
                    headers={
                        'Content-Type': 'application/octet-stream',
                        'Content-MD5': self.fake_image_dict['checksum'],
                    },
                ),
            ]
        )

    def test_download_image_with_fd(self):
        self._register_image_mocks()
        output_file = io.BytesIO()
        self.cloud.download_image(self.image_name, output_file=output_file)
        output_file.seek(0)
        self.assertEqual(output_file.read(), self.output)
        self.assert_calls()

    def test_download_image_with_path(self):
        self._register_image_mocks()
        output_file = tempfile.NamedTemporaryFile()
        self.cloud.download_image(
            self.image_name, output_path=output_file.name
        )
        output_file.seek(0)
        self.assertEqual(output_file.read(), self.output)
        self.assert_calls()

    @mock.patch('openstack.image.v2._proxy.Proxy.find_image')
    def test_get_images(self, mock_find):
        image1 = dict(id='123', name='mickey')
        mock_find.return_value = image1
        r = self.cloud.get_image('mickey')
        self.assertIsNotNone(r)
        self.assertDictEqual(image1, r)

    @mock.patch('openstack.image.v2._proxy.Proxy.find_image')
    def test_get_image_not_found(self, mock_find):
        mock_find.return_value = None
        r = self.cloud.get_image('doesNotExist')
        self.assertIsNone(r)

    def test_get_image_name(self, cloud=None):
        cloud = cloud or self.cloud
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json=self.fake_search_return,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json=self.fake_search_return,
                ),
            ]
        )

        self.assertEqual(self.image_name, cloud.get_image_name(self.image_id))
        self.assertEqual(
            self.image_name, cloud.get_image_name(self.image_name)
        )

        self.assert_calls()

    def test_get_image_by_id(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    json=self.fake_image_dict,
                )
            ]
        )
        self._compare_images(
            self.fake_image_dict, self.cloud.get_image_by_id(self.image_id)
        )
        self.assert_calls()

    def test_get_image_id(self, cloud=None):
        cloud = cloud or self.cloud
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json=self.fake_search_return,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json=self.fake_search_return,
                ),
            ]
        )

        self.assertEqual(self.image_id, cloud.get_image_id(self.image_id))
        self.assertEqual(self.image_id, cloud.get_image_id(self.image_name))

        self.assert_calls()

    def test_get_image_name_operator(self):
        # This should work the same as non-operator, just verifying it does.
        self.test_get_image_name(cloud=self.cloud)

    def test_get_image_id_operator(self):
        # This should work the same as the other test, just verifying it does.
        self.test_get_image_id(cloud=self.cloud)

    def test_empty_list_images(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json={'images': []},
                )
            ]
        )
        self.assertEqual([], self.cloud.list_images())
        self.assert_calls()

    def test_list_images(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json=self.fake_search_return,
                )
            ]
        )
        [
            self._compare_images(a, b)
            for a, b in zip([self.fake_image_dict], self.cloud.list_images())
        ]
        self.assert_calls()

    def test_list_images_show_all(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['member_status=all'],
                    ),
                    json=self.fake_search_return,
                )
            ]
        )
        [
            self._compare_images(a, b)
            for a, b in zip(
                [self.fake_image_dict], self.cloud.list_images(show_all=True)
            )
        ]
        self.assert_calls()

    def test_list_images_show_all_deleted(self):
        deleted_image = self.fake_image_dict.copy()
        deleted_image['status'] = 'deleted'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['member_status=all'],
                    ),
                    json={'images': [self.fake_image_dict, deleted_image]},
                )
            ]
        )
        [
            self._compare_images(a, b)
            for a, b in zip(
                [self.fake_image_dict], self.cloud.list_images(show_all=True)
            )
        ]
        self.assert_calls()

    def test_list_images_no_filter_deleted(self):
        deleted_image = self.fake_image_dict.copy()
        deleted_image['status'] = 'deleted'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json={'images': [self.fake_image_dict, deleted_image]},
                )
            ]
        )
        [
            self._compare_images(a, b)
            for a, b in zip(
                [self.fake_image_dict],
                self.cloud.list_images(filter_deleted=False),
            )
        ]
        self.assert_calls()

    def test_list_images_filter_deleted(self):
        deleted_image = self.fake_image_dict.copy()
        deleted_image['status'] = 'deleted'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json={'images': [self.fake_image_dict, deleted_image]},
                )
            ]
        )
        [
            self._compare_images(a, b)
            for a, b in zip([self.fake_image_dict], self.cloud.list_images())
        ]
        self.assert_calls()

    def test_list_images_string_properties(self):
        image_dict = self.fake_image_dict.copy()
        image_dict['properties'] = 'list,of,properties'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json={'images': [image_dict]},
                ),
            ]
        )
        images = self.cloud.list_images()
        [self._compare_images(a, b) for a, b in zip([image_dict], images)]

        self.assertEqual(
            images[0]['properties']['properties'], 'list,of,properties'
        )
        self.assert_calls()

    def test_list_images_paginated(self):
        marker = str(uuid.uuid4())
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json={
                        'images': [self.fake_image_dict],
                        'next': f'/v2/images?marker={marker}',
                    },
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=[f'marker={marker}'],
                    ),
                    json=self.fake_search_return,
                ),
            ]
        )
        [
            self._compare_images(a, b)
            for a, b in zip([self.fake_image_dict], self.cloud.list_images())
        ]
        self.assert_calls()

    def test_create_image_put_v2_no_import(self):
        self.cloud.image_api_use_tasks = False

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_name],
                        base_url_append='v2',
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['name=' + self.image_name],
                    ),
                    validate=dict(),
                    json={'images': []},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['os_hidden=True'],
                    ),
                    json={'images': []},
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json=self.fake_image_dict,
                    validate=dict(
                        json={
                            'container_format': 'bare',
                            'disk_format': 'qcow2',
                            'name': self.image_name,
                            'owner_specified.openstack.md5': self.fake_image_dict[  # noqa: E501
                                'owner_specified.openstack.md5'
                            ],
                            'owner_specified.openstack.object': self.object_name,  # noqa: E501
                            'owner_specified.openstack.sha256': self.fake_image_dict[  # noqa: E501
                                'owner_specified.openstack.sha256'
                            ],
                            'visibility': 'private',
                            'tags': ['tag1', 'tag2'],
                        }
                    ),
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id, 'file'],
                        base_url_append='v2',
                    ),
                    request_headers={
                        'Content-Type': 'application/octet-stream'
                    },
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.fake_image_dict['id']],
                        base_url_append='v2',
                    ),
                    json=self.fake_image_dict,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.fake_image_dict['id']],
                        base_url_append='v2',
                    ),
                    complete_qs=True,
                    json=self.fake_image_dict,
                ),
            ]
        )

        self.cloud.create_image(
            self.image_name,
            self.imagefile.name,
            wait=True,
            timeout=1,
            tags=['tag1', 'tag2'],
            is_public=False,
            validate_checksum=True,
        )

        self.assert_calls()
        self.assertEqual(
            self.adapter.request_history[7].text.read(), self.output
        )

    def test_create_image_put_v2_import_supported(self):
        self.cloud.image_api_use_tasks = False

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_name],
                        base_url_append='v2',
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['name=' + self.image_name],
                    ),
                    validate=dict(),
                    json={'images': []},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['os_hidden=True'],
                    ),
                    json={'images': []},
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json=self.fake_image_dict,
                    headers={
                        'OpenStack-image-import-methods': IMPORT_METHODS,
                    },
                    validate=dict(
                        json={
                            'container_format': 'bare',
                            'disk_format': 'qcow2',
                            'name': self.image_name,
                            'owner_specified.openstack.md5': self.fake_image_dict[  # noqa: E501
                                'owner_specified.openstack.md5'
                            ],
                            'owner_specified.openstack.object': self.object_name,  # noqa: E501
                            'owner_specified.openstack.sha256': self.fake_image_dict[  # noqa: E501
                                'owner_specified.openstack.sha256'
                            ],
                            'visibility': 'private',
                            'tags': ['tag1', 'tag2'],
                        }
                    ),
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id, 'file'],
                        base_url_append='v2',
                    ),
                    request_headers={
                        'Content-Type': 'application/octet-stream'
                    },
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.fake_image_dict['id']],
                        base_url_append='v2',
                    ),
                    json=self.fake_image_dict,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.fake_image_dict['id']],
                        base_url_append='v2',
                    ),
                    complete_qs=True,
                    json=self.fake_image_dict,
                ),
            ]
        )

        self.cloud.create_image(
            self.image_name,
            self.imagefile.name,
            wait=True,
            timeout=1,
            tags=['tag1', 'tag2'],
            is_public=False,
            validate_checksum=True,
        )

        self.assert_calls()
        self.assertEqual(
            self.adapter.request_history[7].text.read(), self.output
        )

    def test_create_image_use_import(self):
        self.cloud.image_api_use_tasks = False

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_name],
                        base_url_append='v2',
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['name=' + self.image_name],
                    ),
                    validate=dict(),
                    json={'images': []},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['os_hidden=True'],
                    ),
                    json={'images': []},
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json=self.fake_image_dict,
                    headers={
                        'OpenStack-image-import-methods': IMPORT_METHODS,
                    },
                    validate=dict(
                        json={
                            'container_format': 'bare',
                            'disk_format': 'qcow2',
                            'name': self.image_name,
                            'owner_specified.openstack.md5': self.fake_image_dict[  # noqa: E501
                                'owner_specified.openstack.md5'
                            ],
                            'owner_specified.openstack.object': self.object_name,  # noqa: E501
                            'owner_specified.openstack.sha256': self.fake_image_dict[  # noqa: E501
                                'owner_specified.openstack.sha256'
                            ],
                            'visibility': 'private',
                            'tags': ['tag1', 'tag2'],
                        }
                    ),
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id, 'stage'],
                        base_url_append='v2',
                    ),
                    request_headers={
                        'Content-Type': 'application/octet-stream'
                    },
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id, 'import'],
                        base_url_append='v2',
                    ),
                    json={'method': {'name': 'glance-direct'}},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.fake_image_dict['id']],
                        base_url_append='v2',
                    ),
                    json=self.fake_image_dict,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.fake_image_dict['id']],
                        base_url_append='v2',
                    ),
                    complete_qs=True,
                    json=self.fake_image_dict,
                ),
            ]
        )

        self.cloud.create_image(
            self.image_name,
            self.imagefile.name,
            wait=True,
            timeout=1,
            tags=['tag1', 'tag2'],
            is_public=False,
            validate_checksum=True,
            use_import=True,
            import_method='glance-direct',
        )

        self.assert_calls()
        self.assertEqual(
            self.adapter.request_history[7].text.read(), self.output
        )

    def test_create_image_task(self):
        self.cloud.image_api_use_tasks = True
        endpoint = self.cloud.object_store.get_endpoint()

        task_id = str(uuid.uuid4())
        args = dict(
            id=task_id,
            status='success',
            type='import',
            result={
                'image_id': self.image_id,
            },
        )

        image_no_checksums = self.fake_image_dict.copy()
        del image_no_checksums['owner_specified.openstack.md5']
        del image_no_checksums['owner_specified.openstack.sha256']
        del image_no_checksums['owner_specified.openstack.object']

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_name],
                        base_url_append='v2',
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['name=' + self.image_name],
                    ),
                    validate=dict(),
                    json={'images': []},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['os_hidden=True'],
                    ),
                    json={'images': []},
                ),
                dict(
                    method='HEAD',
                    uri=f'{endpoint}/{self.container_name}',
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=f'{endpoint}/{self.container_name}',
                    status_code=201,
                    headers={
                        'Date': 'Fri, 16 Dec 2016 18:21:20 GMT',
                        'Content-Length': '0',
                        'Content-Type': 'text/html; charset=UTF-8',
                    },
                ),
                dict(
                    method='HEAD',
                    uri=f'{endpoint}/{self.container_name}',
                    headers={
                        'Content-Length': '0',
                        'X-Container-Object-Count': '0',
                        'Accept-Ranges': 'bytes',
                        'X-Storage-Policy': 'Policy-0',
                        'Date': 'Fri, 16 Dec 2016 18:29:05 GMT',
                        'X-Timestamp': '1481912480.41664',
                        'X-Trans-Id': 'tx60ec128d9dbf44b9add68-0058543271dfw1',
                        'X-Container-Bytes-Used': '0',
                        'Content-Type': 'text/plain; charset=utf-8',
                    },
                ),
                dict(
                    method='GET',
                    # This is explicitly not using get_mock_url because that
                    # gets us a project-id oriented URL.
                    uri='https://object-store.example.com/info',
                    json=dict(
                        swift={'max_file_size': 1000},
                        slo={'min_segment_size': 500},
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=f'{endpoint}/{self.container_name}/{self.image_name}',
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=f'{endpoint}/{self.container_name}/{self.image_name}',
                    status_code=201,
                    validate=dict(
                        headers={
                            'X-Object-Meta-x-sdk-md5': self.fake_image_dict[
                                'owner_specified.openstack.md5'
                            ],
                            'X-Object-Meta-x-sdk-sha256': self.fake_image_dict[
                                'owner_specified.openstack.sha256'
                            ],
                        }
                    ),
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'image', append=['tasks'], base_url_append='v2'
                    ),
                    json={'id': task_id, 'status': 'processing'},
                    validate=dict(
                        json=dict(
                            type='import',
                            input={
                                'import_from': f'{self.container_name}/{self.image_name}',
                                'image_properties': {'name': self.image_name},
                            },
                        )
                    ),
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['tasks', task_id],
                        base_url_append='v2',
                    ),
                    status_code=503,
                    text='Random error',
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['tasks', task_id],
                        base_url_append='v2',
                    ),
                    json=args,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    json=image_no_checksums,
                ),
                dict(
                    method='PATCH',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    validate=dict(
                        json=sorted(
                            [
                                {
                                    'op': 'add',
                                    'value': f'{self.container_name}/{self.image_name}',
                                    'path': '/owner_specified.openstack.object',  # noqa: E501
                                },
                                {
                                    'op': 'add',
                                    'value': self.fake_image_dict[
                                        'owner_specified.openstack.md5'
                                    ],
                                    'path': '/owner_specified.openstack.md5',
                                },
                                {
                                    'op': 'add',
                                    'value': self.fake_image_dict[
                                        'owner_specified.openstack.sha256'
                                    ],
                                    'path': '/owner_specified.openstack.sha256',  # noqa: E501
                                },
                            ],
                            key=operator.itemgetter('path'),
                        ),
                        headers={
                            'Content-Type': 'application/openstack-images-v2.1-json-patch'  # noqa: E501
                        },
                    ),
                    json=self.fake_search_return,
                ),
                dict(
                    method='HEAD',
                    uri=f'{endpoint}/{self.container_name}/{self.image_name}',
                    headers={
                        'X-Timestamp': '1429036140.50253',
                        'X-Trans-Id': 'txbbb825960a3243b49a36f-005a0dadaedfw1',
                        'Content-Length': '1290170880',
                        'Last-Modified': 'Tue, 14 Apr 2015 18:29:01 GMT',
                        'X-Object-Meta-X-Sdk-Sha256': self.fake_image_dict[
                            'owner_specified.openstack.sha256'
                        ],
                        'X-Object-Meta-X-Sdk-Md5': self.fake_image_dict[
                            'owner_specified.openstack.md5'
                        ],
                        'Date': 'Thu, 16 Nov 2017 15:24:30 GMT',
                        'Accept-Ranges': 'bytes',
                        'Content-Type': 'application/octet-stream',
                        'Etag': fakes.NO_MD5,
                    },
                ),
                dict(
                    method='DELETE',
                    uri=f'{endpoint}/{self.container_name}/{self.image_name}',
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    complete_qs=True,
                    json=image_no_checksums,
                ),
            ]
        )

        self.cloud.create_image(
            self.image_name,
            self.imagefile.name,
            wait=True,
            timeout=1,
            disk_format='vhd',
            container_format='ovf',
            is_public=False,
            validate_checksum=True,
            container=self.container_name,
        )

        self.assert_calls()

    def test_delete_autocreated_no_tasks(self):
        self.use_keystone_v3()
        self.cloud.image_api_use_tasks = False
        deleted = self.cloud.delete_autocreated_image_objects(
            container=self.container_name
        )
        self.assertFalse(deleted)
        self.assert_calls([])

    def test_delete_image_task(self):
        self.cloud.image_api_use_tasks = True
        endpoint = self.cloud.object_store.get_endpoint()

        object_path = self.fake_image_dict['owner_specified.openstack.object']

        image_no_checksums = self.fake_image_dict.copy()
        del image_no_checksums['owner_specified.openstack.md5']
        del image_no_checksums['owner_specified.openstack.sha256']
        del image_no_checksums['owner_specified.openstack.object']

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    json=self.fake_image_dict,
                ),
                dict(
                    method='DELETE',
                    uri=f'https://image.example.com/v2/images/{self.image_id}',
                ),
                dict(
                    method='HEAD',
                    uri=f'{endpoint}/{object_path}',
                    headers={
                        'X-Timestamp': '1429036140.50253',
                        'X-Trans-Id': 'txbbb825960a3243b49a36f-005a0dadaedfw1',
                        'Content-Length': '1290170880',
                        'Last-Modified': 'Tue, 14 Apr 2015 18:29:01 GMT',
                        'X-Object-Meta-X-Sdk-Sha256': self.fake_image_dict[
                            'owner_specified.openstack.sha256'
                        ],
                        'X-Object-Meta-X-Sdk-Md5': self.fake_image_dict[
                            'owner_specified.openstack.md5'
                        ],
                        'Date': 'Thu, 16 Nov 2017 15:24:30 GMT',
                        'Accept-Ranges': 'bytes',
                        'Content-Type': 'application/octet-stream',
                        'Etag': fakes.NO_MD5,
                    },
                ),
                dict(
                    method='DELETE',
                    uri=f'{endpoint}/{object_path}',
                ),
            ]
        )

        self.cloud.delete_image(self.image_id)

        self.assert_calls()

    def test_delete_autocreated_image_objects(self):
        self.use_keystone_v3()
        self.cloud.image_api_use_tasks = True
        endpoint = self.cloud.object_store.get_endpoint()
        other_image = self.getUniqueString('no-delete')

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        service_type='object-store',
                        resource=self.container_name,
                        qs_elements=['format=json'],
                    ),
                    json=[
                        {
                            'content_type': 'application/octet-stream',
                            'bytes': 1437258240,
                            'hash': '249219347276c331b87bf1ac2152d9af',
                            'last_modified': '2015-02-16T17:50:05.289600',
                            'name': other_image,
                        },
                        {
                            'content_type': 'application/octet-stream',
                            'bytes': 1290170880,
                            'hash': fakes.NO_MD5,
                            'last_modified': '2015-04-14T18:29:00.502530',
                            'name': self.image_name,
                        },
                    ],
                ),
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        service_type='object-store',
                        resource=self.container_name,
                        append=[other_image],
                    ),
                    headers={
                        'X-Timestamp': '1429036140.50253',
                        'X-Trans-Id': 'txbbb825960a3243b49a36f-005a0dadaedfw1',
                        'Content-Length': '1290170880',
                        'Last-Modified': 'Tue, 14 Apr 2015 18:29:01 GMT',
                        'X-Object-Meta-X-Shade-Sha256': 'does not matter',
                        'X-Object-Meta-X-Shade-Md5': 'does not matter',
                        'Date': 'Thu, 16 Nov 2017 15:24:30 GMT',
                        'Accept-Ranges': 'bytes',
                        'Content-Type': 'application/octet-stream',
                        'Etag': '249219347276c331b87bf1ac2152d9af',
                    },
                ),
                dict(
                    method='HEAD',
                    uri=self.get_mock_url(
                        service_type='object-store',
                        resource=self.container_name,
                        append=[self.image_name],
                    ),
                    headers={
                        'X-Timestamp': '1429036140.50253',
                        'X-Trans-Id': 'txbbb825960a3243b49a36f-005a0dadaedfw1',
                        'Content-Length': '1290170880',
                        'Last-Modified': 'Tue, 14 Apr 2015 18:29:01 GMT',
                        'X-Object-Meta-X-Shade-Sha256': fakes.NO_SHA256,
                        'X-Object-Meta-X-Shade-Md5': fakes.NO_MD5,
                        'Date': 'Thu, 16 Nov 2017 15:24:30 GMT',
                        'Accept-Ranges': 'bytes',
                        'Content-Type': 'application/octet-stream',
                        (
                            'X-Object-Meta-'
                            + self.cloud._OBJECT_AUTOCREATE_KEY
                        ): 'true',
                        'Etag': fakes.NO_MD5,
                        'X-Static-Large-Object': 'false',
                    },
                ),
                dict(
                    method='DELETE',
                    uri=f'{endpoint}/{self.container_name}/{self.image_name}',
                ),
            ]
        )

        deleted = self.cloud.delete_autocreated_image_objects(
            container=self.container_name
        )
        self.assertTrue(deleted)

        self.assert_calls()

    def _image_dict(self, fake_image):
        return self.cloud._normalize_image(meta.obj_to_munch(fake_image))

    def _call_create_image(self, name, **kwargs):
        imagefile = tempfile.NamedTemporaryFile(delete=False)
        imagefile.write(b'\0')
        imagefile.close()
        self.cloud.create_image(
            name,
            imagefile.name,
            wait=True,
            timeout=1,
            is_public=False,
            validate_checksum=True,
            **kwargs,
        )

    def test_create_image_put_v1(self):
        self.cloud.config.config['image_api_version'] = '1'

        args = {
            'name': self.image_name,
            'container_format': 'bare',
            'disk_format': 'qcow2',
            'properties': {
                'owner_specified.openstack.md5': fakes.NO_MD5,
                'owner_specified.openstack.sha256': fakes.NO_SHA256,
                'owner_specified.openstack.object': f'images/{self.image_name}',
                'is_public': False,
            },
        }

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://image.example.com/v1/images/'
                    + self.image_name,
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri='https://image.example.com/v1/images/detail?name='
                    + self.image_name,
                    json={'images': []},
                ),
                dict(
                    method='POST',
                    uri='https://image.example.com/v1/images',
                    json={'image': ret},
                    validate=dict(json=args),
                ),
                dict(
                    method='PUT',
                    uri=f'https://image.example.com/v1/images/{self.image_id}',
                    json=ret,
                    validate=dict(
                        headers={
                            'x-image-meta-checksum': fakes.NO_MD5,
                            'x-glance-registry-purge-props': 'false',
                        }
                    ),
                ),
                dict(
                    method='GET',
                    uri='https://image.example.com/v1/images/detail',
                    json={'images': [ret]},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v1',
                    ),
                    complete_qs=True,
                    json=ret,
                ),
            ]
        )
        self._call_create_image(self.image_name)
        [
            self._compare_images_v1(b, a)
            for a, b in zip(self.cloud.list_images(), [ret])
        ]

    def test_create_image_put_v1_bad_delete(self):
        self.cloud.config.config['image_api_version'] = '1'

        args = {
            'name': self.image_name,
            'container_format': 'bare',
            'disk_format': 'qcow2',
            'properties': {
                'owner_specified.openstack.md5': fakes.NO_MD5,
                'owner_specified.openstack.sha256': fakes.NO_SHA256,
                'owner_specified.openstack.object': f'images/{self.image_name}',
                'is_public': False,
            },
            'validate_checksum': True,
        }

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://image.example.com/v1/images/'
                    + self.image_name,
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri='https://image.example.com/v1/images/detail?name='
                    + self.image_name,
                    json={'images': []},
                ),
                dict(
                    method='POST',
                    uri='https://image.example.com/v1/images',
                    json={'image': ret},
                    validate=dict(json=args),
                ),
                dict(
                    method='PUT',
                    uri=f'https://image.example.com/v1/images/{self.image_id}',
                    status_code=400,
                    validate=dict(
                        headers={
                            'x-image-meta-checksum': fakes.NO_MD5,
                            'x-glance-registry-purge-props': 'false',
                        }
                    ),
                ),
                dict(
                    method='DELETE',
                    uri=f'https://image.example.com/v1/images/{self.image_id}',
                    json={'images': [ret]},
                ),
            ]
        )

        self.assertRaises(
            exceptions.HttpException,
            self._call_create_image,
            self.image_name,
        )

        self.assert_calls()

    def test_update_image_no_patch(self):
        self.cloud.image_api_use_tasks = False

        args = {
            'name': self.image_name,
            'container_format': 'bare',
            'disk_format': 'qcow2',
            'owner_specified.openstack.md5': fakes.NO_MD5,
            'owner_specified.openstack.sha256': fakes.NO_SHA256,
            'owner_specified.openstack.object': f'images/{self.image_name}',
            'visibility': 'private',
        }

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'

        self.cloud.update_image_properties(
            image=image.Image.existing(**ret),
            **{
                'owner_specified.openstack.object': f'images/{self.image_name}'
            },
        )

        self.assert_calls()

    def test_create_image_put_v2_bad_delete(self):
        self.cloud.image_api_use_tasks = False

        args = {
            'name': self.image_name,
            'container_format': 'bare',
            'disk_format': 'qcow2',
            'owner_specified.openstack.md5': fakes.NO_MD5,
            'owner_specified.openstack.sha256': fakes.NO_SHA256,
            'owner_specified.openstack.object': f'images/{self.image_name}',
            'visibility': 'private',
        }

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_name],
                        base_url_append='v2',
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['name=' + self.image_name],
                    ),
                    validate=dict(),
                    json={'images': []},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['os_hidden=True'],
                    ),
                    json={'images': []},
                ),
                dict(
                    method='POST',
                    uri='https://image.example.com/v2/images',
                    json=ret,
                    validate=dict(json=args),
                ),
                dict(
                    method='PUT',
                    uri=f'https://image.example.com/v2/images/{self.image_id}/file',
                    status_code=400,
                    validate=dict(
                        headers={
                            'Content-Type': 'application/octet-stream',
                        },
                    ),
                ),
                dict(
                    method='DELETE',
                    uri=f'https://image.example.com/v2/images/{self.image_id}',
                ),
            ]
        )

        self.assertRaises(
            exceptions.HttpException,
            self._call_create_image,
            self.image_name,
        )

        self.assert_calls()

    def test_create_image_put_v2_wrong_checksum_delete(self):
        self.cloud.image_api_use_tasks = False

        fake_image = self.fake_image_dict

        fake_image['owner_specified.openstack.md5'] = 'a'
        fake_image['owner_specified.openstack.sha256'] = 'b'

        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json=self.fake_image_dict,
                    validate=dict(
                        json={
                            'container_format': 'bare',
                            'disk_format': 'qcow2',
                            'name': self.image_name,
                            'owner_specified.openstack.md5': fake_image[
                                'owner_specified.openstack.md5'
                            ],
                            'owner_specified.openstack.object': self.object_name,  # noqa: E501
                            'owner_specified.openstack.sha256': fake_image[
                                'owner_specified.openstack.sha256'
                            ],
                            'visibility': 'private',
                        }
                    ),
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id, 'file'],
                        base_url_append='v2',
                    ),
                    request_headers={
                        'Content-Type': 'application/octet-stream'
                    },
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.fake_image_dict['id']],
                        base_url_append='v2',
                    ),
                    json=fake_image,
                ),
                dict(
                    method='DELETE',
                    uri=f'https://image.example.com/v2/images/{self.image_id}',
                ),
            ]
        )

        self.assertRaises(
            exceptions.SDKException,
            self.cloud.create_image,
            self.image_name,
            self.imagefile.name,
            is_public=False,
            md5='a',
            sha256='b',
            allow_duplicates=True,
            validate_checksum=True,
        )

        self.assert_calls()

    def test_create_image_put_bad_int(self):
        self.cloud.image_api_use_tasks = False

        self.assertRaises(
            exceptions.SDKException,
            self._call_create_image,
            self.image_name,
            allow_duplicates=True,
            min_disk='fish',
            min_ram=0,
        )

        self.assert_calls()

    def test_create_image_put_user_int(self):
        self.cloud.image_api_use_tasks = False

        args = {
            'name': self.image_name,
            'container_format': 'bare',
            'disk_format': 'qcow2',
            'owner_specified.openstack.md5': fakes.NO_MD5,
            'owner_specified.openstack.sha256': fakes.NO_SHA256,
            'owner_specified.openstack.object': f'images/{self.image_name}',
            'int_v': '12345',
            'visibility': 'private',
            'min_disk': 0,
            'min_ram': 0,
        }

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_name],
                        base_url_append='v2',
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['name=' + self.image_name],
                    ),
                    validate=dict(),
                    json={'images': []},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['os_hidden=True'],
                    ),
                    json={'images': []},
                ),
                dict(
                    method='POST',
                    uri='https://image.example.com/v2/images',
                    json=ret,
                    validate=dict(json=args),
                ),
                dict(
                    method='PUT',
                    uri=f'https://image.example.com/v2/images/{self.image_id}/file',
                    validate=dict(
                        headers={
                            'Content-Type': 'application/octet-stream',
                        },
                    ),
                ),
                dict(
                    method='GET',
                    uri=f'https://image.example.com/v2/images/{self.image_id}',
                    json=ret,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    complete_qs=True,
                    json=ret,
                ),
            ]
        )

        self._call_create_image(
            self.image_name, min_disk='0', min_ram=0, int_v=12345
        )

        self.assert_calls()

    def test_create_image_put_meta_int(self):
        self.cloud.image_api_use_tasks = False

        args = {
            'name': self.image_name,
            'container_format': 'bare',
            'disk_format': 'qcow2',
            'owner_specified.openstack.md5': fakes.NO_MD5,
            'owner_specified.openstack.sha256': fakes.NO_SHA256,
            'owner_specified.openstack.object': f'images/{self.image_name}',
            'int_v': 12345,
            'visibility': 'private',
            'min_disk': 0,
            'min_ram': 0,
        }

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'
        ret['checksum'] = fakes.NO_MD5

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_name],
                        base_url_append='v2',
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['name=' + self.image_name],
                    ),
                    validate=dict(),
                    json={'images': []},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['os_hidden=True'],
                    ),
                    json={'images': []},
                ),
                dict(
                    method='POST',
                    uri='https://image.example.com/v2/images',
                    json=ret,
                    validate=dict(json=args),
                ),
                dict(
                    method='PUT',
                    uri=f'https://image.example.com/v2/images/{self.image_id}/file',
                    validate=dict(
                        headers={
                            'Content-Type': 'application/octet-stream',
                        },
                    ),
                ),
                dict(
                    method='GET',
                    uri=f'https://image.example.com/v2/images/{self.image_id}',
                    json=ret,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    complete_qs=True,
                    json=ret,
                ),
            ]
        )

        self._call_create_image(
            self.image_name, min_disk='0', min_ram=0, meta={'int_v': 12345}
        )

        self.assert_calls()

    def test_create_image_put_protected(self):
        self.cloud.image_api_use_tasks = False

        args = {
            'name': self.image_name,
            'container_format': 'bare',
            'disk_format': 'qcow2',
            'owner_specified.openstack.md5': fakes.NO_MD5,
            'owner_specified.openstack.sha256': fakes.NO_SHA256,
            'owner_specified.openstack.object': f'images/{self.image_name}',
            'int_v': '12345',
            'protected': False,
            'visibility': 'private',
            'min_disk': 0,
            'min_ram': 0,
        }

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'
        ret['checksum'] = fakes.NO_MD5

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_name],
                        base_url_append='v2',
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['name=' + self.image_name],
                    ),
                    validate=dict(),
                    json={'images': []},
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=['os_hidden=True'],
                    ),
                    json={'images': []},
                ),
                dict(
                    method='POST',
                    uri='https://image.example.com/v2/images',
                    json=ret,
                    validate=dict(json=args),
                ),
                dict(
                    method='PUT',
                    uri=f'https://image.example.com/v2/images/{self.image_id}/file',
                    validate=dict(
                        headers={
                            'Content-Type': 'application/octet-stream',
                        },
                    ),
                ),
                dict(
                    method='GET',
                    uri=f'https://image.example.com/v2/images/{self.image_id}',
                    json=ret,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    complete_qs=True,
                    json=ret,
                ),
            ]
        )

        self._call_create_image(
            self.image_name,
            min_disk='0',
            min_ram=0,
            properties={'int_v': 12345},
            is_protected=False,
        )

        self.assert_calls()


class TestImageSuburl(BaseTestImage):
    def setUp(self):
        super().setUp()
        self.os_fixture.use_suburl()
        self.os_fixture.build_tokens()
        self.use_keystone_v3()
        self.use_glance(
            image_version_json='image-version-suburl.json',
            image_discovery_url='https://example.com/image',
        )

    def test_list_images(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json=self.fake_search_return,
                )
            ]
        )
        [
            self._compare_images(b, a)
            for a, b in zip(self.cloud.list_images(), [self.fake_image_dict])
        ]
        self.assert_calls()

    def test_list_images_paginated(self):
        marker = str(uuid.uuid4())
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image', append=['images'], base_url_append='v2'
                    ),
                    json={
                        'images': [self.fake_image_dict],
                        'next': f'/v2/images?marker={marker}',
                    },
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images'],
                        base_url_append='v2',
                        qs_elements=[f'marker={marker}'],
                    ),
                    json=self.fake_search_return,
                ),
            ]
        )
        [
            self._compare_images(b, a)
            for a, b in zip(
                self.cloud.list_images(),
                [self.fake_image_dict, self.fake_image_dict],
            )
        ]
        self.assert_calls()


class TestImageVolume(BaseTestImage):
    def setUp(self):
        super().setUp()
        self.volume_id = str(uuid.uuid4())

    def test_create_image_volume(self):
        self.register_uris(
            [
                self.get_cinder_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'volumev3',
                        append=['volumes', self.volume_id, 'action'],
                    ),
                    json={
                        'os-volume_upload_image': {'image_id': self.image_id}
                    },
                    validate=dict(
                        json={
                            'os-volume_upload_image': {
                                'container_format': 'bare',
                                'disk_format': 'qcow2',
                                'force': False,
                                'image_name': 'fake_image',
                            }
                        }
                    ),
                ),
                # NOTE(notmorgan): Glance discovery happens here, insert the
                # glance discovery mock at this point, DO NOT use the
                # .use_glance() method, that is intended only for use in
                # .setUp
                self.get_glance_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    json=self.fake_image_dict,
                ),
            ]
        )

        self.cloud.create_image(
            'fake_image',
            self.imagefile.name,
            wait=True,
            timeout=1,
            volume={'id': self.volume_id},
        )

        self.assert_calls()

    def test_create_image_volume_duplicate(self):
        self.register_uris(
            [
                self.get_cinder_discovery_mock_dict(),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'volumev3',
                        append=['volumes', self.volume_id, 'action'],
                    ),
                    json={
                        'os-volume_upload_image': {'image_id': self.image_id}
                    },
                    validate=dict(
                        json={
                            'os-volume_upload_image': {
                                'container_format': 'bare',
                                'disk_format': 'qcow2',
                                'force': True,
                                'image_name': 'fake_image',
                            }
                        }
                    ),
                ),
                # NOTE(notmorgan): Glance discovery happens here, insert the
                # glance discovery mock at this point, DO NOT use the
                # .use_glance() method, that is intended only for use in
                # .setUp
                self.get_glance_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'image',
                        append=['images', self.image_id],
                        base_url_append='v2',
                    ),
                    json=self.fake_image_dict,
                ),
            ]
        )

        self.cloud.create_image(
            'fake_image',
            self.imagefile.name,
            wait=True,
            timeout=1,
            volume={'id': self.volume_id},
            allow_duplicates=True,
        )

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import openstack

openstack.enable_logging(debug=True)

cloud = openstack.connect(cloud='my-citycloud', region_name='Buf1')
try:
    server = cloud.create_server(
        'my-server',
        image='Ubuntu 16.04 Xenial Xerus',
        flavor=dict(id='0dab10b5-42a2-438e-be7b-505741a7ffcc'),
        wait=True,
        auto_ip=True,
    )

    print("\n\nFull Server\n\n")
    cloud.pprint(server)

    print("\n\nTurn Detailed Off\n\n")
    cloud.pprint(cloud.get_server('my-server', detailed=False))

    print("\n\nBare Server\n\n")
    cloud.pprint(cloud.get_server('my-server', bare=True))

finally:
    # Delete it - this is a demo
    cloud.delete_server(server, wait=True, delete_ips=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack.tests.unit import base


class TestDomainParams(base.TestCase):
    def get_mock_url(
        self,
        service_type='identity',
        interface='public',
        resource='projects',
        append=None,
        base_url_append='v3',
        qs_elements=None,
    ):
        return super().get_mock_url(
            service_type,
            interface,
            resource,
            append,
            base_url_append,
            qs_elements,
        )

    def test_identity_params_v3(self):
        project_data = self._get_project_data(v3=True)
        self.register_uris(
            [
                # can't retrieve by name
                dict(
                    method='GET',
                    uri=self.get_mock_url(append=[project_data.project_name]),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        qs_elements=[f'name={project_data.project_name}']
                    ),
                    json=dict(
                        projects=[project_data.json_response['project']]
                    ),
                ),
            ]
        )

        ret = self.cloud._get_identity_params(
            domain_id='5678', project=project_data.project_name
        )
        self.assertIn('default_project_id', ret)
        self.assertEqual(ret['default_project_id'], project_data.project_id)
        self.assertIn('domain_id', ret)
        self.assertEqual(ret['domain_id'], '5678')

        self.assert_calls()

    def test_identity_params_v3_no_domain(self):
        project_data = self._get_project_data(v3=True)

        self.assertRaises(
            exceptions.SDKException,
            self.cloud._get_identity_params,
            domain_id=None,
            project=project_data.project_name,
        )

        self.assert_calls()
